#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Modules for RAG.
"""
import torch
import torch.cuda
import torch.nn
import torch.nn.functional as F
from typing import Any, Tuple, Dict, Optional, List, Union, Type

from parlai.agents.hugging_face.t5 import (
    ParlaiT5Encoder,
    ParlaiT5Decoder,
    build_t5,
    set_device,
)
from parlai.agents.transformer.modules import (
    TransformerEncoder,
    TransformerDecoder,
    get_n_positions_from_options,
    create_embeddings,
)
from parlai.core.dict import DictionaryAgent
from parlai.core.opt import Opt
from parlai.core.torch_generator_agent import TorchGeneratorModel
from parlai.utils.torch import padded_tensor

from parlai.agents.rag.retrievers import retriever_factory, Document


class RagModel(TorchGeneratorModel):
    """
    RagModel.

    The RagModel operates in the following phases:
    1) retrieve:    given a tokenized query, return relevant documents
    2) expand:      given queries and documents, expand the inputs n_docs times,
                    concatenating each document with a relevant context
    3) encode:      given expanded input, encode into encoder representations
    4) decoding:    given encoder outputs, compute n_docs decoder representations for
                    each batch item.
    5) marginalize: given the decoded representations, marginalize over the documents
    appropriately.

    The RagModel overloads the `encoder` and `decoder` attributes of your standard
    `TorchGeneratorModel` to accomplish the five phases above.
    """

    def __init__(self, opt, dictionary, retriever_shared=None):
        from parlai.agents.rag.rag import RAG_MODELS

        self.pad_idx = dictionary[dictionary.null_token]
        self.start_idx = dictionary[dictionary.start_token]
        self.end_idx = dictionary[dictionary.end_token]
        super().__init__(self.pad_idx, self.start_idx, self.end_idx)
        self.fp16 = (
            not opt['no_cuda'] and torch.cuda.is_available() and opt.get('fp16', False)
        )
        self.dict = dictionary
        self.embeddings = create_embeddings(
            dictionary, opt['embedding_size'], self.pad_idx
        )
        # attrs
        self.rag_model_type = opt['rag_model_type']
        self.rag_model_interface = RAG_MODELS[self.rag_model_type](opt, self.pad_idx)
        self.generation_model = opt['generation_model']
        self.n_extra_positions = opt['n_extra_positions']
        self.n_positions = get_n_positions_from_options(opt) + opt['n_extra_positions']
        assert opt['n_extra_positions'] >= 0
        self.expanded_input_truncate = min(
            opt['text_truncate'] or opt['truncate'], get_n_positions_from_options(opt)
        )
        if self.n_extra_positions > 0:
            self.expanded_input_truncate = max(
                self.expanded_input_truncate, self.n_extra_positions
            )
        self.min_doc_token_length = opt['min_doc_token_length']

        # modules
        self.retriever = retriever_factory(opt, dictionary, shared=retriever_shared)
        self.seq2seq_encoder = self.build_encoder(
            opt,
            dictionary=dictionary,
            embedding=self.embeddings,
            padding_idx=self.pad_idx,
        )
        self.seq2seq_decoder = self.build_decoder(
            opt, embedding=self.embeddings, padding_idx=self.pad_idx
        )

    @classmethod
    def build_encoder(
        cls,
        opt: Opt,
        *args,
        dictionary: Optional[DictionaryAgent] = None,
        embedding: Optional[torch.nn.Embedding] = None,
        encoder_class: Optional[Type] = None,
        **kwargs,
    ):
        if encoder_class is None:
            assert dictionary is not None
            return RagEncoder(
                opt=opt, dictionary=dictionary, embedding=embedding, **kwargs
            )
        else:
            return encoder_class(opt, *args, **kwargs)

    @classmethod
    def build_decoder(
        cls,
        opt: Opt,
        *args,
        embedding: Optional[torch.nn.Embedding] = None,
        n_positions: Optional[int] = None,
        decoder_class: Optional[Type] = None,
        **kwargs,
    ):
        if decoder_class is None:
            return RagDecoder(opt=opt, embedding=embedding, n_positions=n_positions)
        else:
            return decoder_class(opt, *args, **kwargs)

    def tokenize_query(self, query: str) -> List[int]:
        """
        Tokenize the query for the retriever.
        """
        return self.retriever.tokenize_query(query)

    def get_retriever_delimiter(self) -> str:
        """
        Return the retriever's delimiter.
        """
        return self.retriever.get_delimiter()

    def encoder(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        query_vec: torch.LongTensor,
        input_turns_cnt: torch.LongTensor,
        positions: Optional[torch.LongTensor] = None,
        segments: Optional[torch.LongTensor] = None,
    ) -> Tuple[
        torch.Tensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
        Optional[List[List[Document]]],
        Optional[torch.Tensor],
    ]:
        """
        Retrieve documents and expand input via concatenation.

        Then, encode as usual in the seq2seq encoder.

        :param input:
            2D [bsz, seqlen] input to the encoder
        :param input_lengths:
            1D [bsz] lengths of each input item
        :param query_vec:
            2D [bsz*n_turns, seqlen] input for the retriever
        :param input_turns_cnt:
            1D [bsz] number of dialogue turns for each input example

        :return (encoder_out, encoder_mask, input_turns_cnt, top_docs, top_doc_scores):
            encoder_out: encoded representations of context/document pairs
            encoder_mask: mask for enc_out
            input_turns_cnt: pass along the input turns count for the decoder
            top_docs: List of top Documents for each batch example
            top_doc_scores: scores for each retrieved document.
        """
        # Retrieve, get expanded input
        if all([tensor is not None for tensor in [input_lengths, query_vec]]):
            expanded_input, top_docs, top_doc_scores = self.retrieve_and_concat(
                input, input_lengths, query_vec, input_turns_cnt
            )
        else:
            expanded_input = input
            top_docs = top_doc_scores = None

        # Run through seq2seq encoder
        tensor, mask = self.seq2seq_encoder(
            expanded_input, positions, segments
        )  # type: ignore

        return tensor, mask, input_turns_cnt, top_docs, top_doc_scores

    def decoder(
        self,
        input: torch.LongTensor,
        encoder_state: Tuple[Any, ...],
        incr_state: Optional[Dict[str, Any]] = None,
    ) -> Tuple[torch.Tensor, Optional[Dict[str, Any]]]:
        """
        Decode, RAG-Style.

        Obtain decoder representations as usual, then marginalize appropriately.

        :param input:
            input for the decoder
        :param encoder_state:
            RAG encoder states
        :param incr_state:
            incremental decoder state

        :return (output, new_incr_state):
            return the output token distribution, as well as new incremental state.
        """
        # 1. Get decoder outputs
        enc_out, enc_mask, input_turns_cnt, docs, doc_scores = encoder_state
        dec_out, new_incr_state = self.seq2seq_decoder(
            input, (enc_out, enc_mask), incr_state
        )  # type: ignore
        dec_out = self.decoder_output(dec_out)

        if all([obj is not None for obj in [docs, doc_scores]]):
            # 2. Get logprobs
            n_docs = doc_scores.size(1)
            out_probs = F.log_softmax(
                dec_out, dim=-1, dtype=torch.float32  # type: ignore
            ).view(
                input.shape[0] // n_docs, n_docs, -1, dec_out.size(-1)
            )  # [bsz * beam_size, n_docs, input_len, esz]

            # 3. Marginalize
            marginalized = self.rag_model_interface.marginalize(
                out_probs, F.log_softmax(doc_scores, dim=1), input_turns_cnt
            )
        else:
            # With RAG Sequence Generation, we do not marginalize over documents.
            marginalized = dec_out

        return marginalized, new_incr_state

    def seq2seq_forward_pass(
        self, xs: torch.LongTensor, ys: torch.LongTensor
    ) -> Tuple[torch.Tensor, torch.Tensor, Tuple[Any, ...]]:
        """
        Simulate a standard seq2seq encoder/decoder forward pass.

        Used in thorough decoding.

        :param xs:
            input tokens
        :param ys:
            teacher forced decoder outputs

        :return (logits, preds, encoder_states):
            logits: token output distribution
            preds: max probability token at each output position
            encoder_states: output states from the encoder
        """
        encoder_states = self.seq2seq_encoder(xs)  # type: ignore
        bsz = ys.size(0)
        seqlen = ys.size(1)
        inputs = ys.narrow(1, 0, seqlen - 1)
        dec_inputs = self.rag_model_interface.get_initial_forced_decoder_input(
            bsz,
            inputs,
            n_docs=1,
            start_idx=self.START_IDX,
            end_idx=self.END_IDX,
            input_turns_cnt=None,
        )
        latent, _ = self.seq2seq_decoder(
            dec_inputs, encoder_states, None
        )  # type: ignore
        logits = self.decoder_output(latent)
        _, preds = logits.max(dim=-1)
        return logits, preds, encoder_states

    def decoder_output(self, latent: torch.Tensor) -> torch.Tensor:
        """
        Output layer for the decoder; maps latent state to token distributions.

        :param latent:
            final representations from last decoder layer.

        :return logits:
            return output distribution over tokens.
        """
        return F.linear(latent, self.embeddings.weight)

    def retrieve_and_concat(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        query_vec: torch.LongTensor,
        input_turns_cnt: torch.LongTensor,
    ) -> Tuple[torch.LongTensor, List[List[Document]], torch.Tensor]:
        """
        Retrieve documents, concat with input.

        :param input:
            2D [bsz, seqlen] input to the encoder
        :param input_lengths:
            1D [bsz] lengths of each input item
        :param query_vec:
            2D [bsz*n_turns, seqlen] input for the retriever
        :param input_turns_cnt:
            1D [bsz] number of dialogue turns for each input example

        :return (expanded_input, top_docs, top_doc_scores):
            expanded_input: [bsz * n_docs, seqlen+doc_len] tensor of context/document inputs
            top_docs: List of top documents for each input
            top_doc_scores: document scores for each document
        """
        # 1. Retrieve
        top_docs, top_doc_scores = self.retriever.retrieve(query_vec)

        # 2. Expand the input
        if input_turns_cnt is not None:
            input = input.repeat_interleave(input_turns_cnt, dim=0)  # type: ignore
            input_lengths = input_lengths.repeat_interleave(
                input_turns_cnt, dim=0
            )  # type: ignore
        expanded_input = self.concat_docs_and_input(
            input, input_lengths, top_docs, top_doc_scores.size(1)
        )

        return expanded_input, top_docs, top_doc_scores

    def concat_docs_and_input(
        self,
        input: torch.LongTensor,
        input_lengths: torch.LongTensor,
        top_docs: List[List[Document]],
        max_num_docs: int,
    ) -> torch.LongTensor:
        """
        Add document tokens to input tokens.

        :param input:
            original input tokens
        :param input_lengths:
            original input lengths
        :param top_docs:
            list of n_docs top documents for each input sequence
        :param max_num_docs:
            maximum number of docs out of all examples

        :return (tokens, lengths):
            return expanded token vectors & corresponding lengths
        """
        max_len = self.expanded_input_truncate
        expanded_input = []
        for i, docs in enumerate(top_docs):
            for rank in range(len(docs)):
                input_i = input[i, :]
                doc = docs[rank]
                doc_tokens = self.dict.txt2vec(doc.get_passage_str())
                if self.generation_model == 'bart' and self.n_extra_positions <= 0:
                    # move SOS to start of passage since we append question to end
                    input_i = input_i[1:]
                    sample_doc_tokens = torch.LongTensor(
                        [self.start_idx] + doc_tokens
                    ).to(input)
                else:
                    sample_doc_tokens = torch.LongTensor(doc_tokens).to(input)

                if self.n_extra_positions <= 0:
                    # Prepend document to text
                    input_i_len = input_lengths[i]
                    new_input_length = min(
                        self.expanded_input_truncate - self.min_doc_token_length,
                        input_i_len,
                    )
                    input_i = input_i[input_i_len - new_input_length : input_i_len]
                    doc_max_len = max(max_len - len(input_i), 0)
                    sample_doc_tokens = sample_doc_tokens[:doc_max_len]
                    expanded_input.append(
                        torch.cat([sample_doc_tokens, input_i])[:max_len]
                    )
                else:
                    # Append Document to text
                    sample_doc_tokens = sample_doc_tokens[:max_len]
                    input_i_new = input_i.new(
                        self.n_positions - self.n_extra_positions
                    ).fill_(self.pad_idx)
                    input_i_new[: input_i.size(0)] = input_i
                    expanded_input.append(torch.cat([input_i_new, sample_doc_tokens]))
            # append extra null inputs if there are diff # of docs per input
            expanded_input += [
                input[i, :].new(input[i, :].size()).fill_(self.pad_idx)
            ] * (max_num_docs - len(docs))
        expanded_input, _ = padded_tensor(
            expanded_input,
            fp16friendly=self.fp16,
            max_len=max_len if self.n_extra_positions <= 0 else None,
            pad_idx=self.pad_idx,
        )
        expanded_input = expanded_input.to(input.device)
        return expanded_input  # type: ignore

    def output(self, tensor: torch.Tensor) -> torch.Tensor:
        """
        RAG "output" is already scaled in RagModel.decoder.
        """
        return tensor

    def reorder_encoder_states(
        self,
        encoder_states: Tuple[torch.Tensor, ...],
        indices: Union[List[int], torch.LongTensor],
    ) -> Tuple[torch.Tensor, ...]:
        """
        Reorder the encoder states.

        Each RAG Model type prepares encoder states for generation differently.
        """
        if not torch.is_tensor(indices):
            indices = torch.LongTensor(indices).to(
                encoder_states[0].device
            )  # type: ignore
        return self.rag_model_interface.reorder_encoder_states(encoder_states, indices)

    def reorder_decoder_incremental_state(
        self,
        incremental_state: Dict[str, Any],
        inds: Union[List[int], torch.LongTensor],
    ) -> Optional[Dict[str, Any]]:
        """
        TODO: Determine how to do this
        """
        return None

    def decode_forced(
        self, encoder_states: Tuple[torch.Tensor, ...], ys: torch.LongTensor
    ) -> Tuple[torch.Tensor, torch.LongTensor]:
        """
        Decode with a fixed, true sequence, computing loss.

        Override TGM.decode_forced to both:
        1) handle BART eos/bos issues, and
        2) appropriately get forced decoder input.

        :param encoder_states:
            encoder output states
        :param ys:
            teacher forced label

        :return logits, preds:
            logits: output token distribution (as logits, not probs)
            preds: tokens corresponding with max probs according to output distribution.
        """
        bsz = ys.size(0)
        seqlen = ys.size(1)
        inputs = ys.narrow(1, 0, seqlen - 1)
        if (ys[:, 0] == self.START_IDX).any() and self.generation_model != 'bart':
            raise AssertionError(
                "The Beginning of Sentence token is automatically added to the "
                "label in decode_forced, but you included it in the label. This means "
                "your model will have a double BOS token, which is probably not what "
                "you intended."
            )
        doc_scores = encoder_states[-1]

        inputs = self.rag_model_interface.get_initial_forced_decoder_input(
            bsz,
            inputs,
            n_docs=doc_scores.size(1) if doc_scores is not None else None,
            start_idx=self.START_IDX,
            end_idx=self.END_IDX,
            input_turns_cnt=encoder_states[2],
        )
        latent, _ = self.decoder(inputs, encoder_states)
        logits = self.output(latent)
        _, preds = logits.max(dim=-1)
        return logits, preds  # type: ignore


class RagEncoder(TransformerEncoder):
    """
    Subclass TransformerEncoder to use additional positions if desired.
    """

    def __init__(
        self,
        opt: Opt,
        dictionary: DictionaryAgent,
        embedding: Optional[torch.nn.Embedding] = None,
        padding_idx: int = 0,
    ):
        """
        RagEncoder initialization.

        The Rag Seq2seq encoder is just a regular encoder
        """
        n_init_positions = get_n_positions_from_options(opt) + opt['n_extra_positions']
        super().__init__(
            opt=opt,
            vocabulary_size=len(dictionary),
            embedding=embedding,
            padding_idx=padding_idx,
            reduction_type='none',
            n_positions=n_init_positions,
        )


class RagDecoder(TransformerDecoder):
    """
    RagDecoder is a subclass of TransformerDecoder.

    No further modifications necessary.
    """

    pass


class T5RagModel(RagModel):
    """
    T5 For RAG.
    """

    def __init__(self, opt, dictionary, retriever_shared=None):
        opt['t5'] = build_t5(opt)
        if opt['t5_model_parallel']:
            opt['t5'].parallelize()
        else:
            opt['t5'].deparallelize()
        super().__init__(opt, dictionary, retriever_shared)
        self.embedding_size = opt['t5'].model_dim
        self.t5 = opt.pop('t5', None)

    @classmethod
    def build_encoder(
        cls,
        opt: Opt,
        *args,
        dictionary: Optional[DictionaryAgent] = None,
        embedding: Optional[torch.nn.Embedding] = None,
        encoder_class: Optional[Type] = None,
        **kwargs,
    ):
        return RagModel.build_encoder(
            opt,
            encoder=opt['t5'].get_encoder(),
            encoder_class=ParlaiT5Encoder,
            **kwargs,
        )

    @classmethod
    def build_decoder(
        cls,
        opt: Opt,
        *args,
        embedding: Optional[torch.nn.Embedding] = None,
        n_positions: Optional[int] = None,
        decoder_class: Optional[Type] = None,
        **kwargs,
    ):
        return RagModel.build_decoder(
            opt,
            decoder=opt['t5'].get_decoder(),
            decoder_class=ParlaiT5Decoder,
            **kwargs,
        )

    def reorder_decoder_incremental_state(
        self, incremental_state: Dict[int, dict], inds: torch.Tensor
    ) -> Optional[Dict[int, dict]]:
        return None

    @set_device
    def decoder_output(self, latent: torch.Tensor):
        tensor = latent * (self.t5.model_dim ** -0.5)
        logits = self.t5.lm_head(tensor)
        return logits



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Pre-trained DPR Model.
"""
from parlai.core.build_data import built, download_models, get_model_dir
import os
import os.path

path = 'https://dl.fbaipublicfiles.com/dpr/checkpoint/retriver/multiset'


def download(datapath):
    ddir = os.path.join(get_model_dir(datapath), 'hallucination')
    model_type = 'multiset_dpr'
    version = 'v1.0'
    if not built(os.path.join(ddir, model_type), version):
        opt = {'datapath': datapath, 'model_type': model_type}
        fnames = ['hf_bert_base.cp']
        download_models(
            opt,
            fnames,
            'hallucination',
            version=version,
            use_model_type=True,
            path=path,
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os
from parlai.utils.logging import logger
from parlai.core.build_data import DownloadableFile


MSC_DATASETS_VERSION = 'v0.1'


RESOURCES = [
    DownloadableFile(
        f'http://parl.ai/downloads/msc/msc_{MSC_DATASETS_VERSION}.tar.gz',
        f'msc_{MSC_DATASETS_VERSION}.tar.gz',
        'e640e37cf4317cd09fc02a4cd57ef130a185f23635f4003b0cee341ffcb45e60',
    )
]


def get_msc_dir_path(opt):
    dpath = os.path.join(opt['datapath'], 'msc')
    return dpath


def build(opt):
    version = MSC_DATASETS_VERSION
    # create particular instance of dataset depending on flags..
    dpath = get_msc_dir_path(opt)
    if not build_data.built(dpath, version):
        logger.warning('[build data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)
        # Mark the data as built.
        build_data.mark_done(dpath, version)

    return dpath



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Retrieval-Augmented Generation for Knowledge-Intensive NLP Tasks.

Original Paper: https://arxiv.org/abs/2005.11401

As used in ParlAI: https://arxiv.org/abs/2104.07567
"""
from abc import ABC, abstractstaticmethod
import os
import torch
import torch.nn
import torch.cuda
from typing import Any, Dict, List, Optional, Tuple, Union, Type

from parlai.agents.bart.bart import BartAgent
import parlai.agents.hugging_face.hugging_face  # noqa: F401
from parlai.agents.hugging_face.t5 import T5Agent
from parlai.agents.transformer.polyencoder import PolyencoderAgent
from parlai.agents.transformer.transformer import TransformerGeneratorAgent
from parlai.core.build_data import modelzoo_path
from parlai.core.dict import DictionaryAgent
from parlai.core.message import Message
from parlai.core.metrics import AverageMetric, normalize_answer, F1Metric
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_agent import History, Batch, Output
from parlai.core.torch_generator_agent import PPLMetric, TorchGeneratorAgent, TreeSearch
from parlai.utils.distributed import sync_parameters
from parlai.utils.io import PathManager
import parlai.utils.logging as logging
from parlai.utils.misc import recursive_getattr
import parlai.utils.pickle
from parlai.utils.torch import total_parameters, trainable_parameters, PipelineHelper
from parlai.utils.typing import TShared

from parlai.agents.rag.args import setup_rag_args
from parlai.agents.rag.model_types import (
    RagModelInterface,
    RagSequence,
    RagToken,
    RagTurn,
)
from parlai.agents.rag.modules import RagModel, T5RagModel
from parlai.agents.rag.retrievers import Document


class BaseGenerationAgentMixin(ABC):
    """
    A Base Generation Agent Mixin.
    """

    @abstractstaticmethod
    def build_rag_model(opt: Opt, dictionary: DictionaryAgent) -> RagModel:
        """
        Build and return a RAG Model.
        """


class TransformerGeneratorRagAgent(TransformerGeneratorAgent, BaseGenerationAgentMixin):
    @staticmethod
    def build_rag_model(opt: Opt, dictionary: DictionaryAgent) -> RagModel:
        return RagModel(opt, dictionary)


class BartRagAgent(BartAgent, BaseGenerationAgentMixin):
    @staticmethod
    def build_rag_model(opt: Opt, dictionary: DictionaryAgent) -> RagModel:
        return RagModel(opt, dictionary)


class T5RagAgent(T5Agent, BaseGenerationAgentMixin):
    @staticmethod
    def build_rag_model(opt: Opt, dictionary: DictionaryAgent) -> T5RagModel:
        return T5RagModel(opt, dictionary)

    def _generate(
        self,
        batch: Batch,
        beam_size: int,
        max_ts: int,
        prefix_tokens: Optional[torch.LongTensor] = None,
    ) -> Tuple[List[Tuple[torch.LongTensor, torch.Tensor]], List[TreeSearch]]:
        """
        Override since T5 needs to call TGA generate.
        """
        return TorchGeneratorAgent._generate(
            self, batch, beam_size, max_ts, prefix_tokens
        )


GENERATION_AGENTS = {
    'transformer/generator': TransformerGeneratorRagAgent,
    'bart': BartRagAgent,
    't5': T5RagAgent,
}

RAG_MODELS = {'sequence': RagSequence, 'token': RagToken, 'turn': RagTurn}


class RagAgent(TransformerGeneratorRagAgent, BartRagAgent, T5RagAgent):
    """
    RagAgent.

    The RAG Agent interacts with the RAG model mostly via it's RAG Model interface.
    """

    _generation_agent: Union[
        Type[T5RagAgent], Type[BartRagAgent], Type[TransformerGeneratorRagAgent]
    ]
    _rag_model_interface: RagModelInterface

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add RAG Args.
        """
        PolyencoderAgent.add_cmdline_args(parser, partial_opt=None)
        TransformerGeneratorRagAgent.add_cmdline_args(parser, partial_opt)
        parser = setup_rag_args(parser)
        RagTurn.add_cmdline_args(parser, partial_opt)
        if partial_opt and partial_opt.get('generation_model') == 'bart':
            BartRagAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        T5RagAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        # BART Agent sets these to True; doesn't let you set anything else
        parser.set_defaults(fp16=False, force_fp16_tokens=False)
        return parser

    @property
    def generation_model(self) -> str:
        return self._generation_model

    @generation_model.setter
    def generation_model(self, model: str):
        self._generation_model = model
        self._generation_agent = GENERATION_AGENTS[model]

    @property
    def rag_model_type(self) -> str:
        return self._rag_model_type

    @rag_model_type.setter
    def rag_model_type(self, model: str):
        self._rag_model_type = model
        self._rag_model_interface = RAG_MODELS[model](self.opt, self.NULL_IDX)

    @property
    def retriever_query(self) -> str:
        return self._retriever_query

    @retriever_query.setter
    def retriever_query(self, query: str):
        self._retriever_query = query
        if query == 'one_turn':
            self._query_key = 'text'
        elif query == 'full_history':
            self._query_key = 'full_text'

    def __init__(self, opt: Opt, shared: TShared = None):
        self.opt = opt
        if torch.cuda.is_available():
            torch.cuda.set_device('cuda:0')
        self.generation_model = opt['generation_model']
        self.regret = opt['regret']
        self.regret_intermediate_maxlen = opt['regret_intermediate_maxlen']
        self.retriever_query = opt['rag_retriever_query']

        # Super call
        self._generation_agent.__init__(self, opt, shared)  # type: ignore
        self.rag_model_type = opt['rag_model_type']

        if not shared and self.regret:
            self.regret_model = self.build_regret_model()
        elif shared:
            self.regret_model = shared.get('regret_model')

    def build_regret_model(self) -> RagModel:
        """
        Build and return regret RagModel.
        """
        model_file = modelzoo_path(self.opt['datapath'], self.opt['regret_model_file'])
        if model_file:
            assert os.path.exists(
                model_file
            ), f'specify correct path for --regret-model-file (currently {model_file})'
            regret_opt = Opt.load(f'{model_file}.opt')
            regret_opt['n_docs'] = self.opt['n_docs']  # Urgent that this is the same
            # add keys that were not in this model when originally trained
            regret_opt.update(
                {k: v for k, v in self.opt.items() if k not in regret_opt}
            )
            retriever_shared = None
            if all(
                [
                    regret_opt[k] == self.opt[k]
                    for k in [
                        'rag_retriever_type',
                        'path_to_index',
                        'path_to_dpr_passages',
                    ]
                ]
            ):
                logging.warning('Sharing retrievers between model and regret model!')
                retriever_shared = self.model.retriever.share()
            elif self.opt['regret_override_index']:
                # Sharing Index Path & Passages only; not the full retriever
                logging.warning('Overriding initial ReGReT model index')
                regret_opt['path_to_index'] = self.opt['path_to_index']
                regret_opt['path_to_dpr_passages'] = self.opt['path_to_dpr_passages']

            if self.opt['regret_dict_file']:
                regret_opt['dict_file'] = self.opt['regret_dict_file']

            regret_dict = self.dictionary_class()(regret_opt)
            model = RagModel(regret_opt, regret_dict, retriever_shared=retriever_shared)
            with PathManager.open(model_file, 'rb') as f:
                states = torch.load(
                    f,
                    map_location=lambda cpu, _: cpu,
                    pickle_module=parlai.utils.pickle,
                )
            assert 'model' in states
            model.load_state_dict(states['model'])
            if self.model_parallel:
                ph = PipelineHelper()
                ph.check_compatibility(self.opt)
                model = ph.make_parallel(model)
            elif self.use_cuda:
                model.cuda()
            if self.fp16:
                model = model.half()

            sync_parameters(model)
            train_params = trainable_parameters(model)
            total_params = total_parameters(model)
            logging.info(
                f"Total regret parameters: {total_params:,d} ({train_params:,d} trainable)"
            )
        else:
            model = self.model

        return model

    def share(self):
        shared = super().share()
        if self.regret:
            shared['regret_model'] = self.regret_model
        return shared

    ########################################
    # BART-Specific Overrides              #
    ########################################

    def _convert_model(self, opt: Opt) -> Opt:
        """
        Override BartAgent._convert_model to use RagConversionScript.
        """
        return self._generation_agent._convert_model(self, opt)  # type: ignore

    def _initialize_bart(self, opt: Opt) -> Opt:
        return self._generation_agent._initialize_bart(self, opt)  # type: ignore

    ########################################
    # TorchGeneratorAgent Overrides        #
    ########################################

    ###### 0. Act/Observe ######
    def observe(self, observation: Union[Dict, Message]) -> Message:
        """
        Overrides TA.observe to tokenize the retriever query.
        """
        observation = Message(observation)
        observation = self._generation_agent.observe(self, observation)
        if observation.is_padding():
            return observation
        if 'query_vec' not in observation:
            self._set_query_vec(observation)
        if 'input_turn_cnt_vec' not in observation:
            self._set_input_turn_cnt_vec(observation)
        return observation

    def eval_step(self, batch: Batch) -> Optional[Output]:
        output = super().eval_step(batch)
        if output is None or not hasattr(self.model, 'retriever'):
            return output
        assert isinstance(self.model, RagModel)
        if hasattr(self.model.retriever, 'top_docs'):
            output.top_docs = self.model.retriever.top_docs  # type: ignore
        return output

    ###### 1. Model Inputs ######

    def _model_input(
        self, batch: Batch
    ) -> Tuple[torch.LongTensor, torch.LongTensor, torch.LongTensor, torch.LongTensor]:
        """
        Get input for RagModel.

        :param batch:
            batch to process

        :return (text_vec, text_lengths, query_vec, input_turn_cnt_vec):
            text_vec - tokenized batch
            text_lengths - length of each item in the batch
            query_vec - tokenized vectors for retrieval mechanism
            input_turn_cnt_vec - count of input turns for each batch item
        """
        return (
            batch.text_vec,
            batch.text_vec.ne(self.NULL_IDX).sum(1),
            batch.query_vec,
            batch.input_turn_cnt_vec,
        )

    def _encoder_input(
        self, batch: Batch
    ) -> Tuple[
        torch.LongTensor,
        Optional[torch.LongTensor],
        Optional[torch.LongTensor],
        Optional[torch.LongTensor],
    ]:
        """
        Called directly when generating.

        Some RAG Model Types require different encoder inputs.

        :param batch:
            batch to process

        :return (text_vec, text_lengths, query_vec, input_turn_cnt_vec):
            text_vec - tokenized batch
            text_lengths - length of each item in the batch
            query_vec - tokenized vectors for retrieval mechanism
            input_turn_cnt_vec - count of input turns for each batch item
        """
        if hasattr(self._rag_model_interface, 'get_generation_input'):
            return self._rag_model_interface.get_generation_input(batch)  # type: ignore
        return self._model_input(batch)

    ##### 2. Standard TGA Function Overrides #####
    def build_model(self) -> RagModel:
        """
        Build and return RagModel.
        """
        model = self._generation_agent.build_rag_model(self.opt, self.dict)
        if self.opt['embedding_type'] != 'random':
            self._copy_embeddings(
                model.encoder.embeddings.weight, self.opt['embedding_type']
            )
        return model

    def _resize_token_embeddings(self, state_dict, msg=None):
        """
        Resize the token embeddings when are adding extra special tokens.

        Modify TGA._resize_token_embeddings to access correct modules within RAG.
        """
        # map extra special tokens carefully
        new_size = self.model.embeddings.weight.size()[0]
        orig_size = state_dict['embeddings.weight'].size()[0]
        logging.info(f'Resizing token embeddings from {orig_size} to {new_size}')
        if new_size <= orig_size:
            # new size should be greater than original size,
            # as we are adding special tokens
            raise RuntimeError(msg)

        for emb_weights in [
            'embeddings.weight',
            'seq2seq_encoder.embeddings.weight',
            'seq2seq_decoder.embeddings.weight',
        ]:
            # get new_embs
            old_embs = state_dict[emb_weights]
            new_embs = recursive_getattr(self.model, emb_weights).to(old_embs.device)
            # copy over old weights
            new_embs.data[:orig_size, :] = old_embs.data[:orig_size, :]
            # reset in state dict
            state_dict[emb_weights] = new_embs

        return state_dict

    def build_dictionary(self) -> DictionaryAgent:
        """
        Build and return dictionary.
        """
        return self._generation_agent.build_dictionary(self)

    @staticmethod
    def update_state_dict(
        opt: Opt, state_dict: Dict[str, torch.Tensor], model: torch.nn.Module
    ):
        """
        Update the given state dict to be RAG-ified.

        :param opt:
            options
        :param state_dict:
            weights to load
        :param model:
            underlying model that will load the state_dict

        :return updated_state_dict:
            return state_dict with appropriate keys/values
        """
        # 1. Substitute all "encoder" and "decoder" keys with "seq2seq_encoder" and "seq2seq_decoder"
        if not [k for k in state_dict if k.startswith('seq2seq')]:
            for k in list(state_dict.keys()):
                if k.startswith('encoder') or k.startswith('decoder'):
                    weights = state_dict.pop(k)
                    state_dict[f'seq2seq_{k}'] = weights
        # 2. Retriever state
        if not [k for k in state_dict if 'retriever' in k]:
            retriever_state = {
                f"retriever.{k}": v
                for k, v in model.retriever.state_dict().items()  # type: ignore
            }
            state_dict.update(retriever_state)
        # 3. Handle n_positional difference
        if opt.get('n_extra_positions', 0) > 0:
            key = 'seq2seq_encoder.position_embeddings.weight'
            init_weight = (
                model.seq2seq_encoder.position_embeddings.weight  # type: ignore
            )
            if state_dict[key].size(0) < opt['n_positions'] + opt['n_extra_positions']:
                # Make sure we're not adding more positions to a model trained
                # with extra positions
                state_dict[key] = torch.cat(
                    [
                        state_dict[key].to(init_weight),  # type: ignore
                        init_weight[-opt['n_extra_positions'] :, :],  # type: ignore
                    ],
                    dim=0,
                )
        return state_dict

    def _should_override_dpr_model_weights(self, opt: Opt):
        """
        Determine if we need to override the DPR Model weights.

        Under certain circumstances, one may wish to specify a different `--dpr-model-
        file` for a pre-trained, RAG model. Thus, we additionally check to make sure
        that the loaded DPR model weights are not overwritten by the state loading.
        """
        override_dpr = False
        overrides = opt.get('override', {})
        if overrides.get('dpr_model_file') and os.path.exists(
            overrides['dpr_model_file']
        ):
            override_dpr = True
            logging.warning(
                f"Overriding DPR Model with {modelzoo_path(opt['datapath'], opt['dpr_model_file'])}"
            )
        return override_dpr

    def load_state_dict(self, state_dict: Dict[str, torch.Tensor]):
        """
        Potentially update state dict with relevant RAG components.

        Useful when initializing from a normal seq2seq model.
        """
        try:
            if self._should_override_dpr_model_weights(self.opt):
                state_dict.update(
                    {
                        f"retriever.{k}": v
                        for k, v in self.model.retriever.state_dict().items()  # type: ignore
                    }
                )
            self.model.load_state_dict(state_dict)
        except RuntimeError as msg:
            state_dict = self.update_state_dict(self.opt, state_dict, self.model)
            msg_ = str(msg)
            if 'size mismatch' in msg_ and 'embedding' in msg_:
                if hasattr(self, 'special_toks') and len(self.special_toks) > 0:
                    state_dict = self._resize_token_embeddings(state_dict, msg_)
                    self.resized_embeddings = True  # make note that we resized here
                else:
                    raise RuntimeError(
                        f'{msg_}\n'
                        '-----------------\n'
                        'Could not load the model due to a size mismatch in the '
                        'embeddings. A common reason for this is trying to load '
                        'a model trained with fp16 but loaded without fp16. Try '
                        'adding --fp16 true or --force-fp16-tokens true.'
                    )
            self.model.load_state_dict(state_dict)

    def batchify(self, obs_batch: List[Message], sort: bool = False) -> Batch:
        """
        Override TA.batchify to incorporate query and input turn vecs.
        """
        assert not sort
        if len(obs_batch) == 0:
            return Batch(batchsize=0)

        valid_exs = [ex for ex in obs_batch if self.is_valid(ex)]

        if len(valid_exs) == 0:
            return Batch(batchsize=0)

        batch = self._generation_agent.batchify(self, obs_batch, sort)

        if any(ex.get('query_vec') is not None for ex in valid_exs):
            _qs = []
            for ex in valid_exs:
                q = ex.get('query_vec', self.EMPTY)
                if type(q) is list and type(q[0]) is list:
                    # handling input turns
                    _qs += q
                else:
                    _qs.append(q)
            self.set_batch_query(batch, _qs)

        batch.input_turn_cnt_vec = None
        if any(ex.get('input_turn_cnt_vec') is not None for ex in valid_exs):
            batch.input_turn_cnt_vec = torch.cat(
                [
                    ex.get('input_turn_cnt_vec', torch.LongTensor([1]))
                    for ex in valid_exs
                ],
                dim=0,
            )
        return batch

    def vectorize(self, *args, **kwargs):
        """
        Override TA.vectorize to ensure proper super class handles vectorization.
        """
        return self._generation_agent.vectorize(self, *args, **kwargs)

    def _set_text_vec(
        self, obs: Message, history: History, truncate: Optional[int]
    ) -> Message:
        """
        Override TA._set_text_vec to ensure proper gen agent handles _set_text_vec.
        """
        return self._generation_agent._set_text_vec(self, obs, history, truncate)

    ##### 3. Custom Batch handling #####

    def set_batch_query(self, batch: Batch, queries: List[torch.LongTensor]) -> Batch:
        """
        Put the queries in the batch.

        :param batch:
            batch to put queries in
        :param queries:
            list of query tokens, presumably

        :return batch:
            return the batch, with queries.
        """
        qs, q_lens = self._pad_tensor(queries)
        batch.query_vec = qs
        batch.query_lengths = torch.LongTensor(q_lens)
        return batch

    def _set_query_vec(self, observation: Message) -> Message:
        """
        Tokenize the query for retrieval.

        :param observation:
            observation with input text.

        :return observation:
            return observation with query vec.
        """
        query_str = observation[self._query_key]
        if hasattr(self.model, 'module'):
            observation['query_vec'] = self.model.module.tokenize_query(query_str)
        else:
            observation['query_vec'] = self.model.tokenize_query(query_str)
        return observation

    def _set_input_turn_cnt_vec(self, observation: Message) -> Message:
        """
        Compute the number of turns of context for the observation query.

        :param observation:
            observation with input text.

        :return observation:
            return observation with input turn vec.
        """
        observation = self._rag_model_interface.set_input_turn_cnt_vec(
            observation, self.model, observation[self._query_key]
        )
        return observation

    ##### 4. Decoder Input Functions #####

    def _get_initial_decoder_input(
        self, bsz: int, beam_size: int, dev: torch.device
    ) -> torch.LongTensor:
        """
        Override TGA._get_initial_decoder_input to seed decoder with proper amount of
        inputs.

        This is called during generation, so additional inputs are necessary only for
        RAG Token (which retrieves during the generation step).
        """
        dec_input = self._generation_agent._get_initial_decoder_input(
            self, bsz, beam_size, dev
        )
        rag_dec_input = self._rag_model_interface.get_initial_decoder_input(dec_input)
        return rag_dec_input

    def _get_next_decoder_input(
        self,
        prev_input: torch.LongTensor,
        selection: torch.LongTensor,
        incr_state_inds: torch.LongTensor,
    ) -> torch.LongTensor:
        """
        Override TGA._get_next_decoder_input to repeat decoder input appropriately.
        """
        if hasattr(self._rag_model_interface, 'get_next_decoder_input'):
            dec_input = self._rag_model_interface.get_next_decoder_input(  # type: ignore
                prev_input, selection, incr_state_inds
            )
        else:
            dec_input = self._generation_agent._get_next_decoder_input(
                self, prev_input, selection, incr_state_inds
            )
        return dec_input

    ####################################
    # Model Generation Overrides       #
    ####################################

    ##### Standard TGA Generation Overrides #####

    def _get_context(self, batch: Batch, batch_idx: int) -> torch.LongTensor:
        """
        Override TGA._get_context for rag-sequence/turn models.

        Reason: the batchsize is artificially higher (n_docs * batchsize)
        """
        if hasattr(self._rag_model_interface, 'get_ctxt_index'):
            batch_idx = self._rag_model_interface.get_ctxt_index(  # type: ignore
                batch, batch_idx
            )
        return self._generation_agent._get_context(self, batch, batch_idx)

    def _generate(
        self,
        batch: Batch,
        beam_size: int,
        max_ts: int,
        prefix_tokens: Optional[torch.LongTensor] = None,
    ) -> Tuple[List[Tuple[torch.LongTensor, torch.Tensor]], List[TreeSearch]]:
        """
        Override TGA._generate to potentially call ReGReT.

        TGA._generate is implemented in _rag_generate
        """
        if self.regret:
            beam_preds_scores, _ = self._regret_generate(
                batch, beam_size, self.regret_intermediate_maxlen, prefix_tokens
            )
            preds, _ = zip(*beam_preds_scores)
            new_batch = self._regret_rebatchify(batch, preds)  # type: ignore
            gen_outs = self._rag_generate(new_batch, beam_size, max_ts, prefix_tokens)
        else:
            gen_outs = self._rag_generate(batch, beam_size, max_ts, prefix_tokens)

        return gen_outs

    def _rerank_beams(
        self,
        batch: Batch,
        n_best_beam_preds_scores: List[List[Tuple[torch.LongTensor, torch.Tensor]]],
    ) -> List[List[Tuple[torch.LongTensor, torch.Tensor]]]:
        """
        Optional rerank beams, according to RAG Model type.

        :param batch:
            current batch
        :param n_best_beam_preds_scores:
            bsz-length list of Tuples of predictions and scores

        :return List((pred, score)):
            return a re-ranked version of the n_best_beam_preds_scores
        """
        return self._rag_model_interface.rerank_beams(
            self.model, batch, n_best_beam_preds_scores
        )

    ##### RAG-Specific Generation Functions #####

    def _rag_generate(
        self,
        batch: Batch,
        beam_size: int,
        max_ts: int,
        prefix_tokens: Optional[torch.LongTensor] = None,
    ) -> Tuple[List[Tuple[torch.LongTensor, torch.Tensor]], List[TreeSearch]]:
        """
        Separate from _generate to handle regret.
        """
        batch = self._rag_model_interface.augment_batch_for_generation(
            batch, self.model
        )
        return self._generation_agent._generate(
            self, batch, beam_size, max_ts, prefix_tokens
        )

    def _regret_generate(
        self,
        batch: Batch,
        beam_size: int,
        max_ts: int,
        prefix_tokens: Optional[torch.LongTensor] = None,
    ):
        """
        Swap generation model prior to full generation.
        """
        model, model_type = self.model, self.rag_model_type
        self.model = self.regret_model
        self.rag_model_type = self.regret_model.rag_model_type

        outs = self._rag_generate(batch, beam_size, max_ts, prefix_tokens)

        self.model = model
        self.rag_model_type = model_type
        return outs

    def _regret_rebatchify(
        self, batch: Batch, pred_vecs: List[torch.LongTensor]
    ) -> Batch:
        """
        Create a new batch for the model.

        Given a model's prediction/generation, create a new batch with this text substituted in as the query.

        :param batch:
            prior model batch
        :param pred_vecs:
            predicted, tokenized text from initial batch

        :return batch:
            return a new batch for the model.
        """
        new_queries: List[torch.LongTensor] = []
        query_lens: torch.LongTensor = batch.query_lengths
        query_vec: torch.LongTensor = batch.query_vec

        for i in range(batch.batchsize):
            vec_i = pred_vecs[i]
            txt_i = self._v2t(vec_i)
            query_i = torch.LongTensor(self.model.tokenize_query(txt_i))
            if self.retriever_query == 'one_turn':
                new_queries.append(query_i)
            else:
                query_i = torch.cat([query_vec[i][: query_lens[i]], query_i], dim=0)
                new_queries.append(query_i)  # type: ignore

        self.set_batch_query(batch, new_queries)

        if (
            batch.input_turn_cnt_vec is not None
            and self.retriever_query == 'full_history'
        ):
            batch.input_turn_cnt_vec += 1

        return batch

    ########################################
    # Computing Loss / Recording Metrics   #
    ########################################

    def _record_retrieval_metrics(self, batch: Batch, encoder_state: Tuple[Any, ...]):
        """
        Compute retrieval metrics, given retrieved documents.

        Only works when `--debug` is set.

        If there is knowledge in the Batch, we compute the following metrics:
        A) Doc Level:
        1. recall @ 1 --> is the correct document the first document?
        2. recall @ N --> is the correct document in the first N docs?

        B) Passage Level:
        1. recall @ 1 --> is the correct passage in the first document?
        2. recall @ N --> is the correct passage in the first N docs?

        Only works in debug mode.

        :param batch:
            training/eval batch
        :param encoder_state:
            encoder states from RagEncoder
        """
        if batch.valid_indices is None or batch.observations is None:
            return
        docs: List[List[Document]] = []
        _, _, input_turns_cnt, docs, _ = encoder_state
        if input_turns_cnt is not None:
            new_docs = []
            offset = 0
            for it in input_turns_cnt:
                docs_it = [dd for d in docs[offset : offset + it] for dd in d]
                new_docs.append(docs_it)
                offset += it
            docs = new_docs
        title_key = self.opt['gold_knowledge_title_key']
        passage_key = self.opt['gold_knowledge_passage_key']
        batchsize = len(batch.valid_indices)
        n_docs = self.opt['n_docs']
        metrics = {
            k: [0] * batchsize
            for k in [
                'doc_r@1',
                f'doc_r@{n_docs}',
                'passage_r@1',
                f'passage_r@{n_docs}',
                'title@1_f1',
                'passage@1_f1',
            ]
        }
        for i in range(batchsize):
            ex = batch.observations[i]
            label_title = normalize_answer(ex.get(title_key, ''))
            label_passage = normalize_answer(ex.get(passage_key, ''))

            for rank, doc in enumerate(docs[i]):
                model_title = normalize_answer(doc.get_title())
                model_passage = normalize_answer(doc.get_text())

                title_exact_match = model_title == label_title
                passage_match = (
                    model_passage in label_passage or label_passage in model_passage
                )

                if rank == 0:
                    metrics['doc_r@1'][i] = int(title_exact_match)
                    metrics['passage_r@1'][i] = int(passage_match)
                    metrics['title@1_f1'][i] = F1Metric.compute(
                        guess=model_title, answers=[label_title]
                    ).value()
                    metrics['passage@1_f1'][i] = F1Metric.compute(
                        guess=model_passage, answers=[label_passage]
                    ).value()
                metrics[f'doc_r@{n_docs}'][i] = int(
                    metrics[f'doc_r@{n_docs}'][i] or title_exact_match
                )
                metrics[f'passage_r@{n_docs}'][i] = int(
                    metrics[f'passage_r@{n_docs}'][i] or passage_match
                )

        for m in metrics:
            self.record_local_metric(m, AverageMetric.many(metrics[m], [1] * batchsize))

    def get_model_output(self, batch: Batch) -> Tuple[Any, ...]:
        """
        Return model output.

        :param batch:
            batch to process

        :return model_output:
            return output from model
        """
        if not self.regret:
            model_output = self.model(
                *self._model_input(batch), ys=batch.label_vec
            )  # type: ignore
            scores, preds, enc_state, *_ = model_output
        else:
            with torch.no_grad():
                beam_preds_scores, beams = self._regret_generate(
                    batch, self.beam_size, self.regret_intermediate_maxlen
                )
            regret_preds, _ = zip(*beam_preds_scores)
            new_batch = self._regret_rebatchify(batch, regret_preds)  # type: ignore
            regret_model_output = self.model(
                *self._model_input(new_batch), ys=batch.label_vec
            )  # type: ignore
            regret_scores, preds, enc_state = regret_model_output
            scores = regret_scores

        return (scores, preds, enc_state)

    def compute_loss(
        self, batch: Batch, return_output: bool = False
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, Any]]:
        """
        Override standard TGA.compute_loss to call relevant RAG Model Interface.
        """
        if batch.label_vec is None:
            raise ValueError('Cannot compute loss without a label.')

        model_output = self.get_model_output(batch)
        scores, preds, enc_state, *_ = model_output

        self._record_retrieval_metrics(batch, enc_state)
        (
            loss,
            metric_loss,
            metric_correct,
            metric_target_tokens,
        ) = self._rag_model_interface.compute_loss(
            self.criterion, scores, preds, enc_state, batch.label_vec
        )

        self.record_local_metric(
            'loss', AverageMetric.many(metric_loss, metric_target_tokens)
        )
        self.record_local_metric(
            'ppl', PPLMetric.many(metric_loss, metric_target_tokens)
        )
        self.record_local_metric(
            'token_acc', AverageMetric.many(metric_correct, metric_target_tokens)
        )
        self.record_local_metric(
            'token_em',
            AverageMetric.many(
                [x == y for x, y in zip(metric_correct, metric_target_tokens)]
            ),
        )

        if return_output:
            return loss, model_output
        else:
            return loss

    def _construct_token_losses(self, labels, model_output):
        """
        Override default `_construct_token_loss` so that sequence model does not throw
        an error while being evaled with the `--verbose` flag.

        TODO: implement this
        """
        return None



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
List of all opt presets distributed with ParlAI.

This file is for automatically generating docs.
"""

# PRESET_DESCRIPTIONS is a dictionary mapping alias names to human descriptions
# for the sake of documentation
PRESET_DESCRIPTIONS = {
    "gen/meena": (
        "Inference parameters for the Sample & Rank procedure of Meena. "
        "See [Adiwardana et al. (2020)](https://arxiv.org/abs/2001.09977)."
    ),
    "arch/blenderbot_3B": (
        "Architecture parameters (number layers, etc) for BlenderBot 3B. See "
        "[Roller et al. (2020)](https://arxiv.org/abs/2004.13637)"
    ),
    "gen/blenderbot": (
        "Beam search parameters for BlenderBot. See"
        "[Roller et al. (2020)](https://arxiv.org/abs/2004.13637)"
    ),
    "arch/bart_large": (
        "Architecture parameters (number layers, etc.) for BART-Large. See "
        "[Lewis et. al. (2019](https://arxiv.org/abs/1910.13461)"
    ),
}



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Wrappers for GPT models from HF (in ParlAI).

Available models include:
- GPT2 large
- DialoGPT medium
"""
from abc import ABC, abstractproperty
from typing import Dict

from parlai.core.agents import create_agent
from projects.safety_bench.utils.wrapper_loading import register_model_wrapper


class GPTWrapper(ABC):
    """
    Base class wrapper for GPT wrapper
    """

    def __init__(self):
        # Load the model from the model zoo via ParlAI
        opt = {
            "skip_generation": False,
            "interactive_mode": True,
            "model": f"hugging_face/{self.model_name}",
            "gpt2_size": self.model_size,
            "add_special_tokens": False,
        }
        opt.update(self.additional_opts)
        self.model = create_agent(opt)

    @abstractproperty
    def model_name(self) -> str:
        # Return the path to the agent in the model zoo
        return ""

    @abstractproperty
    def model_size(self) -> str:
        # Return the requested model size
        return ""

    @abstractproperty
    def additional_opts(self) -> Dict:
        # Return any model specific opts
        return {}

    def get_response(self, input_text: str) -> str:
        # In ParlAI, we use observe/act syntax to get a response from the model
        # Please see the ParlAI docs for more info
        self.model.observe({"text": input_text, "episode_done": True})
        response = self.model.act()

        return response.get("text")


@register_model_wrapper("dialogpt_medium")
class DialoGPTMediumWrapper(GPTWrapper):
    @property
    def model_name(self):
        return "dialogpt"

    @property
    def model_size(self):
        return "medium"

    @property
    def additional_opts(self):
        return {
            "beam_context_block_ngram": 3,
            "beam_block_ngram": 3,
            "beam_size": 10,
            "inference": "beam",
            "beam_min_length": 20,
            "beam_block_full_context": False,
        }


@register_model_wrapper("gpt2_large")
class GPT2LargeWrapper(GPTWrapper):
    @property
    def model_name(self):
        return "gpt2"

    @property
    def model_size(self):
        return "large"

    @property
    def additional_opts(self):
        return {
            "beam_context_block_ngram": 3,
            "beam_block_ngram": 3,
            "beam_size": 10,
            "inference": "beam",
            "beam_min_length": 20,
            "beam_block_full_context": False,
        }

    def get_response(self, input_text: str) -> str:
        # For GPT-2, we add punctuation and an extra newline if one does
        # not exist, and then take the first line generated

        if input_text.strip()[-1] not in ['.', '?', '!']:
            input_text += "."

        self.model.observe({"text": input_text + "\n", "episode_done": True})
        response = self.model.act()
        # split on newline
        response_texts = response.get("text").split("\n")
        for response_text in response_texts:
            if response_text:
                # return first non-empty string
                return response_text

        # produced only newlines or empty strings
        return ""
